"""
Note envelope segmentation methods.
"""
import librosa
import numpy as np

import iracema.features
import iracema.spectral
import iracema.pitch

from iracema.core.point import Point
from iracema.core.segment import Segment


def segment_notes(audio, onsets, window=1024, hop=441):
    """
    Extract the note envelope for each note in the excerpt.

    Estimate the following points, whitin each two adjacent onsets:
    end of attack, start of release, offset.

    Arguments
    ---------
    audio : ir.Audio
        Audio object.
    onsets : ir.PointList
        Point list containing the instants of note onset.
    window : int
        Window length.
    hop : int
        Hop length.

    Return
    ------
    note_list: list
        A list of dictionaries with the keys 'onset', 'attack_end',
        'release_start' and 'offset', containing the points that
        correspond to those instants.
    """
    stft = iracema.spectral.STFT(audio, window, hop)
    pitch = iracema.pitch.crepe(audio, min_confidence=0)
    pitch.pad(0, stft.nsamples - pitch.nsamples)

    pitch_midi = pitch.copy()
    pitch_midi.data = librosa.hz_to_midi(pitch.data)
    pitch_midi.unit = 'note'
    pitch_diff = pitch_midi.diff()
    rms = iracema.features.rms(audio, window, hop)
    spf_diff = iracema.features.spectral_flux(stft)

    note_list = []
    for onset_0, onset_1 in zip(onsets[0:-1], onsets[1:]):
        ioi = Segment(onset_0, onset_1)
        this_spf_dif = spf_diff[ioi]

        # release
        release_start = (
            onset_0 + (np.argmax(this_spf_dif.data) / this_spf_dif.fs)
        )

        this_rms = rms[ioi]
        this_pitch_diff = pitch_diff[ioi]

        release_start_point = Point(release_start)

        # attack
        seg_onset_release = Segment(onset_0, release_start)
        rms_onset_release = this_rms[seg_onset_release]
        if rms_onset_release.nsamples <= 1:
            attack_end = release_start
        else:
            attack_end = (
                onset_0 + (np.argmax(rms_onset_release.data) / this_rms.fs)
            )

        attack_end_point = Point(attack_end)

        # offset
        seg_release_end_ioi = Segment(release_start, onset_1)
        pitch_diff_end = this_pitch_diff[seg_release_end_ioi]

        pitch_change_idxs = np.where(pitch_diff_end.data > 0.5)[0]
        if not np.any(pitch_change_idxs):
            offset = onset_1
        else:
            idx = pitch_change_idxs[0] - 1
            offset = release_start +(idx / this_pitch_diff.fs)

        offset_point = Point(offset)

        # save points to dictionary list
        points = [attack_end_point, release_start_point]
        if offset_point != onset_1:
            points.append(offset_point)

        note = {
            'onset': onset_0,
            'attack_end': attack_end_point,
            'release_start': release_start_point,
            'offset': offset_point
        }

        note_list.append(note)
    
    return note_list



"""
Note onset detection methods.
"""
import warnings 
from pathlib import Path

import numpy as np
import scipy.signal as sig
from tensorflow.keras.models import load_model

import iracema.features
import iracema.pitch
import iracema.core.segment
import iracema.core.point
from iracema.plot import waveform_trio_features_and_points
from iracema.segmentation.odfs import (odf_rms_derivative, odf_pitch_change,
                                       odf_adaptative_rms)
from iracema.util.ml import activations_to_points
from .util import convolve_activations, three_sliced_mel_spectrograms


def cnn_model(
    audio,
    instrument='clarinet',
    smooth_odf=True,
    odf_threshold = 0.328,
    display_plot = False,
    return_odf_data = False,
):
    """
    Extract the note onsets using the CNN method.

    Arguments
    ---------
    audio: ir.Audio
        Audio file to be processed.
    instrument: string
        Name of the instrument (currently trained only for clarinet).
    smooth_odf: bool
        If true, the final ODF will be smoothed by convolving it with a hanning
        window of length 5.
    odf_threshold : float
        Minimum threshold for the peak picking in the ODF curve.
    display_plot : bool
        Whether of not to plot the results
    return_odf_data : bool
        Whether or not to return the odf data

    """
    audio_ = audio.copy()
    if float(audio_.fs) != 44100.:
        audio_ = audio_.resample(44100)

    if instrument != 'clarinet':
        warnings.warn(
            "This model is specialized in clarinet recordings. It might not "
            "perform well for other instruments."
        )

    window, hop = 1024, 441
    sliced_spectrogram, frame_fs = three_sliced_mel_spectrograms(
        audio_,
        window,
        hop,
        frames_per_slice = 15,
        n_mels = 80,
        fmin = 27.5,
        fmax = 16000,
        db = True,
    )

    here = Path(__file__).parent
    model_file = here/'clari-onsets.h5'
    model = load_model(model_file)
    y_pred = model.predict(sliced_spectrogram)
    if smooth_odf:
        y_pred[:, 0] = convolve_activations(y_pred[:, 0])

    onsets = activations_to_points(
        y_pred,
        frame_fs,
        encoding='binary',
        threshold=odf_threshold,
        peak_pick=True,
    )
    odf_data = y_pred[:, 0]

    if display_plot:
        waveform_trio_features_and_points(audio_, odf_data, onsets)

    if return_odf_data:
        return onsets, odf_data
    return onsets


def adaptative_rms(
        audio,
        short_window=512,
        long_window=4096,
        hop=512,
        alpha=0.1,
        min_time=None,
        odf_threshold=0.2,
        display_plot=False,
        display_plot_rms=False,
        return_odf_data=False,
):
    """
    Extract the note onsets using the adaptative RMS method.

    Arguments
    ---------
    audio : Audio
        Audio time series.
    short_window : int
        Length of the short term window for the calculation of the RMS.
    long_window : int
        Length of the long term window for the calculation of the RMS.
    hop : int
        Length of the hop for the sliding window.
    alpha : float
        Reduction factor for the long term RMS curve.
    display_plot_rms : bool
        Whether of not to plot the RMS curves used to calculate the ODF.
    min_time : float
        Minimum time (in seconds) between successive onsets.
    odf_threshold : float
        Ratio of the ODF maxima to be defined as a minimum threshold
        for the peak picking.
    display_plot : bool
        Whether of not to plot the results.
    return_odf_data : bool
        Whether or not to return the odf data.

    Return
    ------
    onsets : PointList
        List of onsets.
    odf_data : TimeSeries
        Time series containing the onset detection function obtained. This will
        only be returned if the argument `return_odf_data` has been set to
        True.
    """
    onsets, odf_data = extract_from_odf(
        audio,
        odf_adaptative_rms,
        long_window=long_window,
        short_window=short_window,
        hop=hop,
        alpha=alpha,
        display_plot_rms=display_plot_rms,
        min_time=min_time,
        odf_threshold=odf_threshold,
        odf_threshold_criteria="relative_to_max",
        display_plot=display_plot,
    )

    if return_odf_data:
        return onsets, odf_data
    return onsets


def rms_derivative(
        audio,
        window=1024,
        hop=512,
        min_time=None,
        odf_threshold=0.2,
        display_plot=False,
        return_odf_data=False,
):
    """
    Extract note onsets from the ``audio`` time-series using its ``rms``.
    The RMS will be calculated if it's not passed as an argument. The argument
    ``min_time`` can be used to specify the minimum distance (in seconds)
    between two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    window : int
        Window length for computing the RMS.
    hop : int
        Hop length for computing the RMS.
    min_time : float, optional
        Minimum time (in seconds) between successive onsets.
    odf_threshold : float
        Minimum threshold for the peak picking in the ODF curve.
    display_plot : bool
        Whether of not to plot the results
    return_odf_data : bool
        Whether or not to return the odf data

    Return
    ------
    onsets : list
        List of onset points.
    odf_data : TimeSeries
        Time series containing the onset detection function obtained. This will
        only be returned if the argument `return_odf_data` has been set to
        True.
    """
    onsets, odf_data = extract_from_odf(
        audio,
        odf_rms_derivative,
        window=window,
        hop=hop,
        min_time=min_time,
        odf_threshold=odf_threshold,
        display_plot=display_plot,
    )

    if return_odf_data:
        return onsets, odf_data
    return onsets


def pitch_variation(
        audio,
        window,
        hop,
        minf0=120,
        maxf0=4000,
        smooth_pitch=True,
        min_time=None,
        odf_threshold=0.04,
        display_plot=False,
        return_odf_data=False,
):
    """
    Extract note onsets from the ``audio`` time-series using its ``pitch``.
    The argument ``min_time`` can be used to specify the minimum distance (in
    seconds) between two adjacent onsets.

    Args
    ----
    audio : Audio
        Audio object
    window : int
        Window length for computing the pitch.
    hop : int
        Hop length for computing the pitch.
    minf0 : int
        Minimum frequency for the pitch detection.
    maxf0 : int
        Maximum frequency for the pitch detection.
    smooth_pitch: bool
        Whether or not the pitch curve should be smoothed.
    min_time : float
        Minimum time (in seconds) between successive onsets.
    odf_threshold : float
        Minimum threshold for the peak picking in the ODF curve
    display_plot : bool
        Whether of not to plot the results
    return_odf_data : bool
        Whether or not to return the odf data

    Return
    ------
    onsets : list
        List of onset points.
    odf_data : TimeSeries
        Time series containing the onset detection function obtained. This will
        only be returned if the argument `return_odf_data` has been set to
        True.
    """
    odf = odf_pitch_change

    onsets, odf_data = extract_from_odf(
        audio,
        odf,
        window=window,
        hop=hop,
        min_time=min_time,
        odf_threshold=odf_threshold,
        minf0=minf0,
        maxf0=maxf0,
        smooth_pitch=smooth_pitch,
        display_plot=display_plot,
    )

    if return_odf_data:
        return onsets, odf_data
    return onsets


def extract_from_odf(
        audio,
        odf,
        min_time=None,
        odf_threshold=0.2,
        odf_threshold_criteria="absolute",
        display_plot=False,
        **parameters,
):
    """
    Generic method to extract onsets from an ODF (onset detection function).

    Arguments
    ---------
    audio : Audio
        Audio time series.
    odf : function
        Reference to the ODF.
    min_time : float
        Minimum time (in seconds) between successive onsets.
    odf_threshold : float
        Minimum ODF threshold for a peak to be considered as an onset.
    odf_threshold_criteria : string ['absolute', 'relative_to_max']
        Specifies how the argument ``odf_threshold`` will be used: if
        ``'absolute'`` its value will be used directly as the threshold;
        else, if ``'relative_to_max'``, its value will be used to calculate
        the threshold, relative to the maximum value in the ODF curve, e.g.:
        ``odf_threshold``==``0.2`` set the threshold to 20% of the maximum
        value of the ODF curve.
    display_plot : bool
        Whether or not to plot the results.

    Return
    ------
    onsets: PointList
        List of onsets.
    odf_data: TimeSeries
        Time series containing the onset detection function obtained.
    """
    odf_data = odf(audio, **parameters)

    if min_time:
        min_dist = int(min_time * odf_data.fs)
        if min_dist == 0:
            min_dist = None
    else:
        min_dist = None

    if odf_threshold_criteria == "absolute":
        threshold = odf_threshold
    elif odf_threshold_criteria == "relative_to_max":
        threshold = odf_threshold * np.max(odf_data.data)
    else:
        raise ValueError(
            ("Invalid value for argument `odf_threshold_criteria`: "
             f"'{odf_threshold_criteria}'"))

    ixs_onsets, _ = sig.find_peaks(
        odf_data.data, height=threshold, distance=min_dist)

    onsets = iracema.core.point.PointList([
        iracema.core.point.Point.from_sample_index(ix, odf_data)
        for ix in ixs_onsets
    ])

    if display_plot:
        waveform_trio_features_and_points(audio, odf_data, onsets)

    return onsets, odf



"""
This method implements the configuration loading funcionalities for iracema.
"""

import os
from dataclasses import dataclass, field
from pathlib import Path


@dataclass
class ConfigLoader():
    """
    Configuration loader class.
    """
    samples_path: str
    _samples_path: str = field(init=False, repr=False)

    @property
    def samples_path(self):  # pylint: disable=function-redefined
        """
        Path to the directory where the sample audio files should be located.
        """
        return self._samples_path

    @samples_path.setter
    def samples_path(self, path: str):
        self._samples_path = os.path.abspath(Path(path))

    @classmethod
    def get_default_config_loader(cls, iracema_root):
        """
        Factory method to instantiate a default config object.
        """
        relative_samples_path = Path('..', 'audio', 'iracema-audio')
        return cls(Path(iracema_root / relative_samples_path))



import numpy as np


def evaluate_onsets(targets, predictions, tolerance=0.01):
    """
    Evaluation metrics for onset detection methods.

    Arguments
    ---------
    target : numpy array
        Target onsets (annotations).
    prediction : numpy array
        Predicted onsets (to be evaluated).
    tolerance : float
        Maximum time tolerance in seconds.

    Returns
    -------
    metrics : dict
        A dictionary with the calculated metrics.
    """
    # generate distances matrix between targets and predictions
    mesh = np.array(np.meshgrid(targets, predictions))
    combinations = mesh.T.reshape(-1, len(predictions), 2)
    distances = np.abs(np.diff(combinations, axis=2)).squeeze()

    # search for best matches within the tolerance distance
    target_found = np.full(len(targets), np.nan)
    while True:
        ix_min_dist = np.unravel_index(
            np.argmin(distances, axis=None), distances.shape)
        distance = distances[ix_min_dist]
        if np.isnan(target_found[ix_min_dist[0]]):
            target_found[ix_min_dist[0]] = ix_min_dist[1]
        distances[ix_min_dist] = np.inf
        if distance > tolerance:
            break

    arange_predictions = np.arange(0, len(predictions))
    true_positives = np.sum(~np.isnan(target_found))
    false_negatives = np.sum(np.isnan(target_found))
    false_positives = len(
        np.setdiff1d(
            arange_predictions,
            target_found[~np.isnan(target_found)],
            assume_unique=True))

    precision = true_positives / (true_positives + false_positives)
    recall = true_positives / (true_positives + false_negatives)
    fmeasure = (2 * precision * recall) / (precision + recall)

    metrics = {
        'true_positives': true_positives,
        'false_negatives': false_negatives,
        'false_positives': false_positives,
        'precision': precision,
        'recall': recall,
        'f-measure': fmeasure,
    }

    return metrics



"""
Contains modules that implement the core classes of the project. To make these
classes easily available for users, they are all already imported into the
namespace `iracema`. For instance, the class ``Audio`` is available in the
module ``iracema.core.audio``, but it can be accessed using the shorter name
``iracema.Audio``.
"""



"""
Methods for playing audio.
"""
import sounddevice as sd
import numpy as np

from pkg_resources import resource_filename
from resampy import resample

from iracema.io.audiofile import read


def play(audio_time_series, blocking=False):
    """
    Play audio time series.

    Args
    ----
    audio_time_series: iracema.core.audio.Audio
    blocking: bool
    """
    return _play_stream(audio_time_series, blocking=blocking)


def play_with_clicks(audio_time_series,
                     points,
                     click_file=None,
                     blocking=False,
                     return_time_series=False):
    """
    Play audio with clicks in the instants corresponding to the specified
    points. The path to the audio file containing the click sound must be
    provided.

    Args
    ----
    audio_time_series: iracema.core.audio.Audio
    points: iracema.core.point.PointList
    click_file: string
    blocking: bool
    """
    if not click_file:
        click_file = resource_filename('iracema', 'assets/audio/Click.wav')
    click_sound, fs, _ = read(click_file)
    if fs != audio_time_series.fs:
        click_sound = resample(click_sound, fs, audio_time_series.fs)
    indexes = points.map_indexes(audio_time_series)
    audio_with_clicks = audio_time_series.copy()
    for i in indexes:
        to_pad = i + len(click_sound) - len(audio_with_clicks.data)
        if to_pad > 0:
            audio_with_clicks.data = np.pad(audio_with_clicks.data, (0, to_pad))
        audio_with_clicks.data[i:i + len(click_sound)] += click_sound

    if return_time_series:
        return audio_with_clicks

    return _play_stream(audio_with_clicks, blocking=blocking)


def play_interval_samples(audio_time_series,
                          from_sample,
                          to_sample,
                          soft_start=True,
                          blocking=False):
    """
    Play audio interval from sample `from_sample` to `to_sample`.

    Args
    ----
    audio_time_series: iracema.core.audio.Audio
    from_sample: int
    to_sample: int
    soft_start: bool
    blocking: bool
    """
    data_to_play = audio_time_series.data[from_sample:to_sample]
    size_soft_start = int(audio_time_series.fs * 0.005)
    soft_start = np.concatenate(
        (np.linspace(0, 1, size_soft_start, dtype=np.float_),
         np.ones(data_to_play.shape[-1] - size_soft_start)))

    return _play_stream(soft_start * data_to_play, blocking=blocking)


def play_interval_seconds(audio_time_series,
                          from_seconds,
                          to_seconds,
                          soft_start=True,
                          blocking=False):
    """
    Play audio interval from time `from_seconds` to `to_seconds`.

    Args
    ----
    audio_time_series: iracema.core.audio.Audio
    from_seconds: int
    to_seconds: int
    soft_start: bool
    blocking: bool
    """
    if from_seconds:
        from_sample = int((from_seconds - audio_time_series.start_time) *
                          audio_time_series.fs)
    else:
        from_sample = None
    if to_seconds:
        to_sample = int(
            (to_seconds - audio_time_series.start_time) * audio_time_series.fs)
    else:
        to_sample = None

    play_interval_samples(
        audio_time_series,
        from_sample,
        to_sample,
        soft_start=soft_start,
        blocking=blocking)


def _play_stream(audio_time_series, blocking=False):
    sd.default.blocksize = 256
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            from IPython.display import Audio as IPythonAudio, display
            display(
                IPythonAudio(
                    data=audio_time_series.data,
                    rate=audio_time_series.fs,
                    autoplay=False))
        else:
            sd.play(
                audio_time_series.data,
                audio_time_series.fs,
                blocking=blocking)
            return None
    except NameError:
        sd.play(
            audio_time_series.data, audio_time_series.fs, blocking=blocking)
        return None


def stop():
    """
    Stop playing audio.
    """
    sd.stop()



"""
This module contain classes used to manipulate points in TimeSeries objects.
"""
from collections.abc import MutableSequence
import csv
from decimal import Decimal

import numpy as np

import iracema.core.segment


class Point(Decimal):
    """
    A point object represents an instant in a time series, i.e., one specific
    sample index. It is flexible enough to locate samples corresponding to the
    same instant in time series with different sampling rates.

    .. Hint:: This class is also available at the main package level as
        ``iracema.Point``.
    """
    def __repr__(self):
        return f"Point({self})"

    @classmethod
    def from_sample_index(cls, index, time_series):
        time_offset = Decimal(time_series.start_time)
        time = Decimal(int(index)) / Decimal(int(time_series.fs))
        time += time_offset
        return Point(time)

    @property
    def time(self):
        """
        Return the time of the points.
        This method will be deprecated soon.
        """
        return self

    def map_index(self, time_series):
        time = self - Decimal(time_series.start_time)
        index = time * Decimal(int(time_series.fs))
        return int(round(index))

    def get_value(self, time_series):
        index = self.map_index(time_series)
        return time_series.data[..., index]


class PointList(MutableSequence):
    """
    List of points.

    .. Hint:: This class is also available at the main package level as
        ``iracema.PointList``.
    """
    def __init__(self, points=None):
        super(PointList, self).__init__()
        if (points is not None):
            self._points = list(points)
        else:
            self._points = []

    def __getitem__(self, index):
        if isinstance(index, slice):
            return PointList(self._points[index])
        return self._points[index]

    def __setitem__(self, index, item):
        if not isinstance(item, Point):
            raise ValueError(
                "The list contains an item that is not a ``Point``")
        self._points[index] = item

    def __delitem__(self, index):
        self._points.__delitem__(index)

    def __len__(self):
        return len(self._points)

    def __repr__(self):
        return str(list(self))

    def insert(self, index, item):  # skipcq: PYL-W0221
        if not isinstance(item, Point):
            raise ValueError(
                "The insert item is not a ``Point``")
        return self._points.insert(index, item)

    @classmethod
    def load_from_file(cls, file_name):
        """
        Instantiate a list of points, loaded from a CSV file. Each line in the
        file must contain the position of a single point. The position must be
        specified in `seconds`.
        """
        with open(file_name, 'r', newline='') as f:
            points = cls([
                Point(row[0])
                for row in csv.reader(f, delimiter=',')
            ])
        return points

    def save_to_file(self, file_name):
        "Save list of points to CSV file."
        with open(file_name, 'w', newline='') as f:
            writer = csv.writer(f, delimiter=',')
            for point in self:
                writer.writerow([point])

    @classmethod
    def from_list_of_indexes(cls, list_indexes, time_series):
        """
        Instantiate a list of points from a list of indexes ``list_indexes``
        and a ``time_series`` object.
        """
        return cls([
            Point.from_sample_index(index, time_series)
            for index in list_indexes
        ])

    @property
    def time(self):
        """
        Return a list with the time of the points.
        This method will be deprecated soon.
        """
        return self

    def map_indexes(self, time_series):
        """
        Return an array with the indexes of ``time_series`` that correspond to
        the points in the list.
        """
        return [point.map_index(time_series) for point in self]

    def get_values(self, time_series):
        """
        Get values from the ``time_series`` corresponding to the points in the
        list.
        """
        return [point.get_value(time_series) for point in self]

    def to_numpy(self):
        """
        Retun a numpy array with the time of the points.
        """
        return np.array([np.float(point) for point in self])

    @classmethod
    def from_numpy(cls, array):
        """
        Instantiate a list of points from a numpy array.
        """
        return cls([Point(p) for p in array])

    def to_segments(self):
        """
        Convert list of points to a list of contiguous segments.
        """
        segments =  iracema.core.segment.SegmentList()
        for pt_start, pt_end in zip(self[0:-1], self[1:]):
            print('-- ', pt_start, pt_end)
            segments.append(iracema.core.segment.Segment(pt_start, pt_end))
        return segments
    



"""
Default recipes to run Iracema.
"""


class Run:
    """
    Execute the iracema algorithms for a set of audio files.
    """

    def process_directory(self):
        """
        Load and process all the audio files within the specified directory.
        """

        pass

    def process_file(
            self,
            filepath,
            min_ioi=0.05,
            min_rms=0.005,
            ws=1024,
            hs=256,
            fft_len=None,
            instrument=None,
    ):
        """
        Load and process the specified audio file.

        Arguments
        ---------
        filepath: str
            Path and name of the audio file to be loaded.
        ws: int
            Default window size.
        hs: int
            Default hop size.
        min_ioi: float
            Inter-onset interval threshold.
        min_rms: float
            RMS threshold for the onset detection (avoids background noise).
        instrument: string
            Name of the musical instrument in the recording
        caption: str
            Text description of the file (for plotting purposes)
        """

        pass



"""
This module contain classes used to manipulate segments and slice TimeSeries
objects using them.
"""
from collections.abc import MutableSequence
import csv

import iracema.core.point


class Segment:
    """
    The objects generated from this class are used to retrieve excerpts from
    time-series.

    .. Hint:: This class is also available at the main package level as
        ``iracema.Segment``.
    """

    def __init__(self, start, end):
        """
        Segment

        Args
        ----
        start : ir.Point
            Point corresponding to the start of the segment.
        end : ir.Point
            Point corresponding to the end of the segment.
        """
        if end is not None and start > end:
            raise ValueError("end must be > start")

        self.start = iracema.core.point.Point(start)
        self.end = iracema.core.point.Point(end)

    def __repr__(self):
        "Overload the representation for the class"
        class_name = self.__class__.__name__
        description = f"({self.start}, {self.end})"
        return f"{class_name}{description}"

    def nsamples(self, time_series):
        return (self.end.map_index(time_series) -
                self.start.map_index(time_series))

    @property
    def duration(self):
        return self.end - self.start

    @property
    def start_time(self):
        return self.start

    @property
    def end_time(self):
        return self.end

    def generate_slice(self, time_series):
        """
        Generate a python slice with the sample indexes that correspond to the
        current segment in `time_series`
        """
        slice_start = self.start.map_index(time_series)
        slice_end = self.end.map_index(time_series)
        return slice(slice_start, slice_end)

    def map_indexes(self, time_series):
        """
        Return a tuple with the indexes of ``time_series`` that correspond
        to the segments in the list.
        """
        return (
            self.start.map_index(time_series),
            self.end.map_index(time_series)
        )


class SegmentList(MutableSequence):
    """
    List of segments.

    .. Hint:: This class is also available at the main package level as
        ``iracema.SegmentList``.
    """

    def __init__(self, segments=None):
        super(SegmentList, self).__init__()
        if (segments is not None):
            self._segments = list(segments)
        else:
            self._segments = []

    def __getitem__(self, index):
        if isinstance(index, slice):
            return SegmentList(self._segments[index])
        return self._segments[index]

    def __setitem__(self, index, item):
        if not isinstance(item, Segment):
            raise ValueError(
                "The list contains an item that is not a ``Segment``")
        self._segments[index] = item

    def __delitem__(self, index):
        self._segments.__delitem__(index)

    def __len__(self):
        return len(self._segments)

    def __repr__(self):
        "Overload the representation for the class"
        return str(list(self))

    def insert(self, index, item):  # skipcq: PYL-W0221
        if not isinstance(item, Segment):
            raise ValueError("The insert item is not a ``Segment``")
        return self._segments.insert(index, item)

    def map_indexes(self, time_series):
        """
        Return an array of tuples with the indexes of ``time_series`` that correspond to
        the segments in the list.
        """
        return [seg.map_indexes(time_series) for seg in self]
    
    def add_segment(self, start, end):
        """
        Instantiate a segment with the given `start` and `end` and add to the list.
        """
        self.append(Segment(start, end))

    @classmethod
    def load_from_file(cls, file_name):
        """
        Instantiate a list of segments, loaded from a CSV file. Each line in the
        file must contain the position of a single point. The position must be
        specified in `seconds`.
        """
        with open(file_name, 'r', newline='') as f:
            segments = cls([
                Segment(row[0], row[1])
                for row in csv.reader(f, delimiter=',')
            ])
        return segments

    def save_to_file(self, file_name):
        "Save SegmentList to a CSV file."
        with open(file_name, 'w', newline='') as f:
            writer = csv.writer(f, delimiter=',')
            for segment in self:
                writer.writerow([segment.start, segment.end])
    


"""
Methods for reading and writing raw audio files.
"""
import re

from os.path import basename, expanduser
from tempfile import NamedTemporaryFile
from urllib.request import urlopen

import audioread
import numpy as np
import scipy.io.wavfile

from w3lib.url import canonicalize_url
from w3lib.url import file_uri_to_path

def read(file_location):
    """
    Read audio file from the local file system or download it from URL.

    Arguments
    ---------
    file_location: str
        Path or URL to the file that will be loaded.

    Return
    ------
    data: numpy array
        Audio data.
    fs: int
        Sampling frequency.
    file_name: str
        Base name of the loaded file.
    """
    intmaxabs = 32768.  # maximum value for 16-bit signed integers
    fs = None
    temp_file = None

    if re.match('https://|http://', file_location):
        url = canonicalize_url(file_location)
        response = urlopen(url)
        temp_file = NamedTemporaryFile()
        temp_file.write(response.read())
        file_path = temp_file.name
        file_name = basename(file_uri_to_path(file_location))
    else:
        file_path = expanduser(file_location)
        file_name = basename(file_path)

    with audioread.audio_open(file_path) as input_file:
        fs = input_file.samplerate
        channels = input_file.channels

        # audioread returns buffers containing 16-bit signed integers
        data_int = np.array([], dtype=np.dtype('int16'))
        for frame in input_file:
            frame_int = np.frombuffer(frame, np.dtype('int16'))
            data_int = np.concatenate((data_int, frame_int), axis=0)
        # convert data to float
        data = data_int.astype(
            np.float_, casting='safe')  # pylint: disable=maybe-no-member

        # Conversion to mono (mix both channels)
        if channels > 1:
            data = data.reshape((-1, channels)).T
            data = np.mean(data, axis=0)

        data = data / intmaxabs

    if temp_file:
        temp_file.close()

    return (data, fs, file_name)


def write(filename, data, fs):
    """
    Write the contents of a `data` array to a WAVE file.

    Arguments
    ---------
    filename: str
        Name of the audio file to be loaded.
    data: numpy array
        Data array containing audio samples.
    fs: int
        Sampling frequency.
    """
    scipy.io.wavfile.write(filename, fs, data)



"Useful methods for ML models."
import warnings

import numpy as np
import scipy.signal

import iracema as ir


def points_to_activations(point_list,
                          output_fs,
                          output_length,
                          encoding='binary',
                          warn_lost_onset=True,
                          activation_type='single',
                          triang_value=0.4):
    """
    Convert a point list to an array with the corresponding activations.

    Arguments
    ---------
    point_list : PointList
        List of points
    output_fs : float
        Sampling frequency for the output array.
    output_length : int
        Length of the output array.
    encoding : str
        Type of encoding for the activations:
        - 'binary': a single activation per time step, with 0 representing
          one class and 1 the other.
        - 'onehot': two complementary activations per time step. The first
          correspond to class 1, and the second to class 0.
    warn_lost_onset : bool
        If True, the method will throw a warning whenever an onset is lost due
        to the resolution of the output array. This situation will happen when
        two or more onsets are closer than the resolution of the output array.
    activation_type : str
        Type of activation function to be generated:
        - 'single' will yield single activation points equal to ``1.0`` for the
           given time instants;
        - 'triangular' will yield a triangular activation around the point
           corresponding to each time instant.
    """
    _validate_encoding(encoding)

    time_array = point_list.to_numpy()
    activations = time_array_to_activations(time_array,
                                            output_fs,
                                            output_length,
                                            warn_lost_onset=warn_lost_onset,
                                            activation_type=activation_type,
                                            triang_value=triang_value)
    if encoding == 'onehot':
        activations = binary_to_onehot_activations(activations)
    elif encoding == 'binary':
        activations = np.expand_dims(activations, axis=-1)
    return activations


def activations_to_points(activations, input_fs, encoding='binary', threshold=0.5, peak_pick=False):
    """
    Convert an array with activations to a point list.

    Arguments
    ---------
    activations : np.array
        Activations array.
    encoding : str
        Type of encoding for the activations:
        - 'binary': a single activation per time step, with 0 representing
          one class and 1 the other.
        - 'onehot': two complementary activations per time step. The first
          represent the probability of class 0, and the second the probability
          of class 1. The second values are equivalent to the values produced
          using 'binary' encoding.
    input_fs : float
        Sampling frequency of the activations array.
    threshold : float
        Activation threshold to add a time instant to the output array.
    """
    _validate_encoding(encoding)
    if encoding == 'onehot':
        activations = onehot_to_binary_activations(activations)
    elif encoding == 'binary':
        activations = activations.squeeze()
    time_array = activations_to_time_array(activations, input_fs, threshold=threshold, peak_pick=peak_pick)
    points = ir.PointList.from_numpy(time_array)
    return points


def time_array_to_activations(time_array,
                              output_fs,
                              output_length,
                              warn_lost_onset=True,
                              activation_type='single',
                              triang_value=0.4):
    """
    Convert an array of times to an array of activations.

    Arguments
    ---------
    time_array : np.array
        List of time instants.
    output_fs : float
        Sampling frequency for the output array.
    output_length : int
        Length of the output array.
    warn_lost_onset : bool
        If True, the method will throw a warning whenever an onset is lost due
        to the resolution of the output array. This situation will happen when
        two or more onsets are closer than the resolution of the output array.
    activation_type : str
        Type of activation function to be generated:
        - 'single' will yield single activation points equal to ``1.0`` for the
           given time instants;
        - 'triangular' will yield a triangular activation around the point
           corresponding to each time instant.
        - 'gaussian' will yield a gaussian activation around the point
           corresponding to each time instant.
    """
    _validate_activation_type(activation_type)

    time_indexes = np.round(time_array.astype(float) *
                            float(output_fs)).astype(int)
    activations = np.zeros(output_length)

    if activation_type == 'triangular':
        time_indexes_last = np.clip(time_indexes - 1, 0, None)
        time_indexes_next = np.clip(time_indexes + 1, None, output_length - 1)
        neighbours = np.append(time_indexes_last, time_indexes_next)
        neighbours = np.unique(neighbours)
        activations[neighbours] = triang_value
    elif activation_type == 'gaussian':
        activations = _activations_to_gaussian(activations)

    activations[time_indexes] = 1.

    num_onsets_original = len(time_array)
    num_onsets_final = np.sum(activations == 1.)
    if warn_lost_onset and (num_onsets_final != num_onsets_original):
        onsets_lost = num_onsets_original - num_onsets_final
        onsets_lost_msg = (
            f"Lost {onsets_lost} onsets during conversion from point list to "
            "probability array. The distances between succesive onsets are "
            "probably too small for the sampling frequency being used.")
        warnings.warn(onsets_lost_msg)

    return activations


def activations_to_time_array(activations, input_fs, threshold=0.5, peak_pick=False):
    """
    Convert an array of activations to an array of times.

    Arguments
    ---------
    activations : np.array
        Activations array.
    input_fs : float
        Sampling frequency of the activations array.
    threshold : float
        Activation threshold to add a time instant to the output array.
    """
    if peak_pick:
        activation_indexes, _ = scipy.signal.find_peaks(activations, height=threshold)
        time_array = activation_indexes.astype(float) / float(input_fs)
    else:
        time_array = np.where(activations > threshold)[0] / float(input_fs)
    return time_array


def binary_to_onehot_activations(activations):
    """
    Convert an array containing a single activation per time step to a
    bidimensional array, containing two complementary activations per time
    step: the first corresponds to the probability of non-onset, and the
    second to the probability of onset.
    """
    if activations.ndim != 1:
        raise ValueError(
            "The array `activations` must have only one dimension.")
    activations = np.expand_dims(activations, axis=-1)
    activations = np.repeat(activations, 2, axis=-1)
    activations[..., 1] = 1 - activations[..., 1]  # non-onset
    return activations


def onehot_to_binary_activations(activations):
    """
    Convert a bidimensional array containing two activations per time step to
    an unidimensional array, containing a single activation per time step,
    representing the probability of onset.
    """
    activations = np.squeeze(activations)
    if activations.ndim != 2:
        raise ValueError("The array `activations` must have two dimensions "
                         "with more than one element.")
    activations = activations[..., 0]  # onset
    return activations


def _activations_to_gaussian(activations, std=0.6, length=5):
    g = scipy.signal.windows.gaussian(length, std=std)
    half = length // 2
    activations_gaussian = np.zeros_like(activations)
    for ix in np.where(activations==1.)[0]:
        span = len(activations_gaussian[ix-half:ix-half+length])
        activations_gaussian[ix-half:ix-half+length] = \
            np.clip(activations_gaussian[ix-half:ix-half+length] + g[0:span], None, 1.)
    return activations_gaussian


def _validate_activation_type(activation_type):
    if activation_type not in ('single', 'triangular', 'gaussian'):
        raise ValueError('Invalid value for `activation_type`.')


def _validate_encoding(encoding):
    if encoding not in ('binary', 'onehot'):
        raise ValueError('Invalid value for argument `encoding`.')



"""
Extraction of spectral information.
"""
from decimal import Decimal

import numpy as np
from deprecated.sphinx import deprecated
from librosa.filters import mel
from librosa.core.convert import mel_frequencies
from librosa import cqt, hybrid_cqt, cqt_frequencies

from iracema.util.windowing import apply_sliding_window
from iracema.util import conversion
import iracema.core.timeseries


class STFT(iracema.core.timeseries.TimeSeries):
    "Compute the Short-Time Fourier Transform for the ``time_series``."

    def __init__(self, time_series, window_size, hop_size, fft_len=None):
        """
        Args
        ----
        time_series : TimeSeries
            Time series for applying the STFT.
        window_size : int
        hop_size : int
        fft_len : int
            Length of the FFT. The signal will be zero-padded if ``fft_len`` >
            ``window_size``. The default value is equal to `window_size`.
        """
        if not fft_len:
            fft_len = window_size

        def calculate(x):
            return np.fft.rfft(x, n=fft_len, norm='ortho')

        stft_data = apply_sliding_window(
            time_series.data,
            window_size,
            hop_size,
            calculate,
            window_name='hann')

        new_fs = Decimal(time_series.fs) / Decimal(hop_size)

        super(STFT, self).__init__(
            new_fs,
            data=stft_data,
            start_time=time_series.start_time,
            caption=time_series.caption)

        self.max_frequency = float(time_series.nyquist)
        self.frequencies = np.fft.rfftfreq(fft_len, 1. / float(time_series.fs))

        self.label = 'STFT'
        self.unit = 'complex'

    def magnitude(self, power=2., db=False):
        if float(power) not in (1., 2.):
            raise ValueError(
                'The argument `power` must be equal to 1.0 or 2.0')
        magnitude = np.abs(self.data)**power
        if db:
            if power == 1.0:
                magnitude = conversion.amplitude_to_db(magnitude)
            elif power == 2.0:
                magnitude = conversion.energy_to_db(magnitude)
        return magnitude

    def phase(self):
        return np.angle(self.data)


class Spectrogram(iracema.core.timeseries.TimeSeries):
    "Generate spectrogram for the given `time_series`."

    def __init__(self,
                 time_series,
                 window_size,
                 hop_size,
                 fft_len=None,
                 power=2.,
                 db=False):
        """
        Args
        ----
        time_series : TimeSeries
            Time series for applying the STFT.
        window_size : int
        hop_size : int
        fft_len : int
            Length of the FFT. The signal will be zero-padded if ``fft_len`` >
            ``window_size``. The default value is equal to `window_size`.
        power : float
            Exponent for the spectrogram.
        db : bool
            Whether or not to convert the output values to dB.
        """
        stft = STFT(time_series, window_size, hop_size, fft_len=fft_len)
        data = stft.magnitude(power=power, db=db)

        super(Spectrogram, self).__init__(
            stft.fs,
            data=data,
            start_time=stft.start_time,
            caption=stft.caption)

        self.max_frequency = stft.max_frequency
        self.frequencies = stft.frequencies

        self.label = 'Spectrogram'
        self.unit = 'Magnitude'
        self._power = power
        self._db = db


class MelSpectrogram(iracema.core.timeseries.TimeSeries):
    def __init__(self,
                 time_series,
                 window_size,
                 hop_size,
                 n_mels=256,
                 fft_len=None,
                 power=2.,
                 db=False,
                 fmin=0.,
                 fmax=None,
                 htk=False):
        """
        Compute a mel spectrogram for ``time_series``.

        Args
        ----
        time_series : TimeSeries
            Time series for applying the STFT.
        window_size : int
        hop_size : int
        n_mels : int
            Number of mel-scaled filters/channels.
        fft_len : int
            Length of the FFT. The signal will be zero-padded if ``fft_len`` >
            ``window_size``. The default value is equal to `window_size`.
        power : float
            Exponent for the spectrogram.
        db : bool
            Whether or not to convert the output values to dB.
        fmin : float
            Frequency of the lowest filter.
        fmax : float
            Frequency of the highest filter.
        htk : bool
        """
        if not fft_len:
            fft_len = window_size

        spec = Spectrogram(
            time_series,
            window_size,
            hop_size,
            fft_len=fft_len,
            power=power,
            db=False)

        fmax = fmax or spec.max_frequency
        mel_basis = mel(
            time_series.fs, fft_len, n_mels=n_mels, fmin=fmin, fmax=fmax)
        data = np.dot(mel_basis, spec.data)

        if db:
            if power == 1.0:
                data = conversion.amplitude_to_db(data)
            elif power == 2.0:
                data = conversion.energy_to_db(data)

        super(MelSpectrogram, self).__init__(
            spec.fs,
            data=data,
            start_time=spec.start_time,
            caption=spec.caption)

        self.frequencies = mel_frequencies(
            n_mels=n_mels, fmin=fmin, fmax=fmax, htk=htk)
        self.max_frequency = spec.frequencies[-1]
        self.label = 'Mel Spectrogram'
        self.unit = 'Magnitude'
        self._power = power
        self._db = db


class CQT(iracema.core.timeseries.TimeSeries):
    def __init__(self,
                 time_series,
                 hop_size,
                 n_bins=84,
                 bins_per_octave=12,
                 hybrid=True,
                 power=2.,
                 fmin=None,
                 db=False,
                 **kwargs):
        """
        Compute the Constant Q Transform for ``time_series``.

        Args
        ----
        time_series : TimeSeries
            Time series for applying the STFT.
        window_size : int
        hop_size : int
        fft_len : int
            Length of the FFT. The signal will be zero-padded if ``fft_len`` >
            ``window_size``. The default value is equal to `window_size`.
        power : float
            Exponent for the spectrogram.
        db : bool
            Whether or not to convert the output values to dB.
        n_mels : int
            Number of mel-scaled filters/channels.
        fmin : float
            Frequency of the lowest filter.
        fmax : float
            Frequency of the highest filter.
        htk : bool
        """
        new_fs = Decimal(time_series.fs) / Decimal(hop_size)
        if hybrid:
            cqt_func = hybrid_cqt
        else:
            cqt_func = cqt
        data = np.abs(
            cqt_func(time_series.data,
                sr=float(time_series.fs),
                hop_length=hop_size,
                n_bins=n_bins,
                bins_per_octave=bins_per_octave,
                fmin=fmin,
                **kwargs
            )
        )
        if db:
            if power == 1.0:
                data = conversion.amplitude_to_db(data)
            elif power == 2.0:
                data = conversion.energy_to_db(data)
        frequencies = cqt_frequencies(n_bins,
                                      fmin,
                                      bins_per_octave=bins_per_octave,
                                      tuning=kwargs.get('tuning') or 0)
        super(CQT, self).__init__(
            new_fs,
            data=data,
            start_time=time_series.start_time,
            caption=time_series.caption)

        self.frequencies = frequencies
        self.max_frequency = float(frequencies[-1])

        self.label = 'CQT'
        self.unit = 'Magnitude'
        self._power = power
        self._db = db


@deprecated(version='0.2.0', reason='Deprecated method. Use `STFT` instead.')
def fft(*args, **kwargs):
    "Deprecated FFT method."
    return STFT(*args, **kwargs)



"""
Methods for converting values between different units.
"""
from decimal import Decimal

import numpy as np


def sample_index_to_seconds(sample_index, fs, time_offset=0):
    """
    Convert a sample index to time (in seconds).

    Args
    ----
    sample_index: int
        The index of a sample in a time series.
    fs: float
        Sampling frequency.
    time_offset: float
        Time offset to be added to the result (in seconds).
    """
    sample_index = Decimal(sample_index)
    fs = Decimal(fs)
    time_offset = Decimal(time_offset)
    return sample_index / fs + time_offset


def seconds_to_sample_index(time, fs, time_offset=0):
    """
    Convert time (in seconds) to sample index.

    Args
    ----
    time: float
        Time in seconds.
    fs: float
        Sampling frequency.
    time_offset: float
        Time offset (in seconds) to be subtracted from `time` passed.

    Note
    ----
    The return value will be rounded to an integer number. Therefore, you must
    be careful when doing these conversion operation between sample index and
    time, not to incur in loss of precision.
    """
    return int((time - time_offset) * fs)


def map_sample_index(sample_index, source_fs, source_time_offset, target_fs,
                     target_time_offset):
    """
    Convert a given ``sample_index`` for ``source_fs`` and ``time_offset``
    into the corresponding sample index for ``target_fs`` and
    ``target_time_offset``.

    Args
    ----
    sample_index : int
    source_fs : float
    source_time_offset : float
    target_fs : float
    target_time_offset : float

    Return
    ------
    target_sample_index : int
    """
    seconds = sample_index_to_seconds(
        sample_index, source_fs, time_offset=source_time_offset)

    return seconds_to_sample_index(
        seconds, target_fs, time_offset=target_time_offset)


def amplitude_to_db(amplitude, clip_min=1.e-10):
    """
    Convert amplitude to dB.
    """
    return energy_to_db(amplitude**2, clip_min=clip_min)


def energy_to_db(energy, clip_min=1.e-20):
    """
    Convert energy to dB.
    """
    return 10 * np.log10(np.clip(energy, clip_min, None))


def db_to_amplitude(db):
    """
    Convert dB to amplitude.
    """
    return 10 ** (db / 20)


def db_to_energy(db):
    """
    Convert dB to energy.
    """
    return 10 ** (db / 10)



import matplotlib.pyplot as plt
import numpy as np

import iracema.features
import iracema.pitch
import iracema.spectral
from iracema.aggregation import aggregate_sucessive_samples


def odf_adaptative_rms(audio,
                     long_window=4096,
                     short_window=512,
                     hop=512,
                     alpha=0.1,
                     display_plot_rms=False):
    """
    Arguments
    ---------
    audio: Audio
        Audio time series.
    short_window : int
        Length of the short term window for the calculation of the RMS.
    long_window : int
        Length of the long term window for the calculation of the RMS.
    hop:
        Length of the hop for the sliding window.
    alpha:
        Reduction factor for the long term RMS curve.
    display_plot_rms: bool
        Whether of not to plot the RMS curves.

    Return
    ------
    odf: TimeSeries
        Onset detection function.
    """
    rms_long = iracema.features.rms(audio, long_window, hop)
    rms_short = iracema.features.rms(audio, short_window, hop)\
        .pad_like(rms_long)

    rms_long.data = rms_long.data * (1 - alpha)
    rms_diff = (rms_long - rms_short).hwr()

    odf = rms_diff.copy()
    odf.data = np.zeros_like(rms_diff.data)

    last, sum_, pk, ix_pk = 0, 0, 0, 0
    it = np.nditer(rms_diff.data, flags=['f_index'])
    for x in it:
        if x > 0:
            sum_ += x
            if x > pk:
                pk = x
                ix_pk = it.index
        else:
            if last > 0:
                odf.data[ix_pk] = sum_
            sum_ = 0
            pk = 0
        last = x

    if display_plot_rms:
        plt.plot(rms_long.time, rms_long.data, linewidth=0.5, color='r')
        plt.plot(rms_short.time, rms_short.data, linewidth=0.5, color='b')

    return odf


def odf_rms_derivative(audio, window=1024, hop=512):
    """
    Onset detection function based on RMS.

    Arguments
    ---------
    audio : Audio
        Audio object
    window : int
        Window length for computing the RMS.
    hop : int
        Hop length for computing the RMS.

    Return
    ------
    odf: TimeSeries
        Onset detection function.
    """
    rms = iracema.features.rms(audio, window, hop)
    return rms.diff().hwr()  # * rms


def odf_pitch_change(audio,
                     window=1024,
                     hop=512,
                     minf0=120,
                     maxf0=4000,
                     smooth_pitch=True):
    """
    Onset detection function based on Pitch.

    Arguments
    ---------
    audio : Audio
        Audio object
    window : int
        Window length for computing the pitch.
    hop : int
        Hop length for computing the pitch.
    minf0 : int
        Minimum frequency for the pitch detection.
    maxf0 : int
        Maximum frequency for the pitch detection.
    smooth_pitch: bool
        Whether or not the pitch curve should be smoothed.

    Return
    ------
    odf: TimeSeries
        Onset detection function.
    """
    window_mode = 9
    fft = iracema.spectral.fft(audio, window, hop)
    pitch = iracema.pitch.expan(fft, minf0=minf0, maxf0=maxf0)
    if smooth_pitch:
        pitch = iracema.pitch.pitch_filter(pitch)
        pitch = iracema.pitch.pitch_mode(pitch, window=window_mode)

    def ratio_successive(current, last):
        min_denominator = 0.1
        last = np.float(last) + min_denominator
        ratio = current / last
        if ratio >= 1:
            ratio = ratio - 1
        else:
            ratio = 1 - ratio
        return np.abs(ratio)

    return aggregate_sucessive_samples(pitch, ratio_successive)



"""
Data plotting utilities.
"""
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.widgets import MultiCursor  # pylint: disable=import-error

from iracema.features import rms as rms_, peak_envelope as peak_envelope_
from iracema.util import conversion

DEFAULT_FIG_SIZE = (9, 9)


def line_plot(time_series, linewidth=1, alpha=0.9, figsize=None, **kwargs):
    """
    Plot the time series using matplotlib.
    Line width and alpha values can be set as optional parameters.
    """
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    plt.plot(
        time_series.time,
        time_series.data,
        label=time_series.label,
        linewidth=linewidth,
        alpha=alpha,
        **kwargs)
    if time_series.label:
        plt.legend(loc='lower right', ncol=2, fontsize='x-small')
    plt.title(time_series.caption)
    plt.ylabel(time_series.unit)
    plt.xlabel('time (s)')
    plt.show()

    return f


def spectrogram(spec_ts,
                log=False,
                fmin=0.,
                fmax=None,
                figsize=None,
                normalize=True,
                cmap='viridis'):
    """
    Plot the spectrogram of the audio signal.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    ax = f.subplots(1)

    plt.subplots_adjust(hspace=0.05)

    # plotting spectrogram
    _add_spectrogram_to_axes(
        ax, spec_ts, log, fmin=fmin, fmax=fmax, normalize=normalize, cmap=cmap)

    # show the resulting image
    f.show()

    return f


def waveform_spectrogram(audio,
                         spec_ts,
                         log=False,
                         rms=None,
                         peak_envelope=None,
                         fmin=0.,
                         fmax=None,
                         figsize=None):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    _add_waveform_trio_to_axes(ax1, audio, rms, peak_envelope)

    # plotting spectrogram
    _add_spectrogram_to_axes(ax2, spec_ts, log, fmin=fmin, fmax=fmax)

    # show the resulting image
    f.show()

    return f


def waveform_spectrogram_pitch(audio,
                               spec_ts,
                               pitch,
                               rms=None,
                               peak_envelope=None,
                               log=False,
                               fmin=0.,
                               fmax=None,
                               cmap='viridis',
                               figsize=None):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal and its fundamental frequency  `pitch`.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    _add_waveform_trio_to_axes(ax1, audio, rms, peak_envelope)

    # plotting spectrogram
    _add_spectrogram_to_axes(
        ax2, spec_ts, log, fmin=fmin, fmax=fmax, cmap=cmap)

    # plotting pitch
    _add_curve_to_axes(ax2, pitch, fmt='r', ymin=fmin, ymax=fmax)

    # show the resulting image
    f.show()

    return f


def waveform_spectrogram_harmonics(audio,
                                   spec_ts,
                                   harmonics,
                                   rms=None,
                                   peak_envelope=None,
                                   log=False,
                                   fmin=0.,
                                   fmax=None,
                                   normalize=True,
                                   cmap='viridis',
                                   figsize=None):
    """
    Plot two graphs: the first one showing curves for the ``audio`` waveform,
    the ``rms`` and the ``peak_envelope``; the second showing the spectrogram
    of the audio signal and its `harmonics`.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    (ax1, ax2) = f.subplots(
        2, sharex=True, gridspec_kw={'height_ratios': [1, 3.5]})

    plt.subplots_adjust(hspace=0.05)

    # plotting curves
    _add_waveform_trio_to_axes(
        ax1,
        audio,
        rms,
        peak_envelope,
    )

    # plotting spectrogram
    _add_spectrogram_to_axes(
        ax2,
        spec_ts,
        log,
        fmin=fmin,
        fmax=fmax,
        normalize=normalize,
        cmap=cmap)

    # plotting harmonics
    _add_curve_to_axes(ax2, harmonics, fmt='r', ymin=fmin, ymax=fmax)

    # show the resulting image
    f.show()

    return f


def waveform_and_notes(audio, notes, figsize=None):
    """
    Plot waveform and note segments.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    axes = f.subplots(1)

    _add_waveform_to_axes(axes, audio)
    _add_notes_to_axes(axes, notes)

    f.show()


def waveform_trio(audio, rms=None, peak_envelope=None, figsize=None):
    """
    Plot a graph showing curves for the ``audio`` waveform, the ``rms`` and the
    ``peak_envelope``.
    """
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    plt.figure(figsize=figsize)

    axes = plt.axes()

    # add waveform trio to first axes
    _add_waveform_trio_to_axes(axes, audio, rms, peak_envelope)

    plt.show()


def waveform_trio_and_features(audio,
                               rms=None,
                               peak_envelope=None,
                               features=(),
                               figsize=None):
    """
    Plot a graph showing curves for the ``audio`` waveform, the ``rms`` and the
    ``peak_envelope``; followed by a series of graphs, one for each time-series
    in the tuple `features`.
    """
    if not features:
        raise ValueError("the features to be plotted were not specified")

    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    axes_list = f.subplots(len(features) + 1, sharex=True)

    plt.subplots_adjust(hspace=0.05)

    # add audio to first axes
    _add_waveform_trio_to_axes(axes_list[0], audio, rms, peak_envelope)

    # add features to the other axes
    for i, feature in enumerate(features, start=1):
        _add_curve_to_axes(axes_list[i], feature, label=feature.label)
        axes_list[i].legend(loc='lower right', fontsize='x-small')

    MultiCursor(f.canvas, axes_list, color='gray', lw=1)

    plt.show()

    return f


def waveform_trio_features_and_points(audio,
                                      feature,
                                      point_list,
                                      rms=None,
                                      peak_envelope=None,
                                      figsize=None):
    # configuring figure and subplots
    if not figsize:
        figsize = DEFAULT_FIG_SIZE
    f = plt.figure(figsize=figsize)

    axes_list = f.subplots(2, sharex=True)

    # add waveform trio to first axes
    _add_waveform_trio_to_axes(axes_list[0], audio, rms, peak_envelope)

    # add feature
    _add_curve_to_axes(axes_list[1], feature, label=feature.label)
    axes_list[1].legend(loc='lower right', fontsize='x-small')

    _add_points_to_axes(axes_list[1], point_list.time,
                        point_list.get_values(feature))

    MultiCursor(f.canvas, axes_list, color='gray', lw=1)

    f.show()


def _add_notes_to_axes(axes, notes):
    """
    Add note segments to the given ``axes``.
    """
    for note in notes:
        axes.plot(note.start_time, note.data[0], '^r')
        axes.plot(note.end_time, note.data[-1], 'vr')


def _add_curve_to_axes(axes,
                       time_series,
                       fmt='b',
                       linewidth=1,
                       alpha=0.9,
                       label=None,
                       set_labels=True,
                       ymin=None,
                       ymax=None):
    "Add the curve for the given ``time_series`` to the given ``axes``."

    if set_labels:
        axes.set(ylabel=time_series.unit)
        if time_series.caption:
            axes.set_title(time_series.caption)

    axes.set_xlim(time_series.start_time, time_series.end_time)

    if time_series.nfeatures == 1:
        axes.plot(
            time_series.time,
            time_series.data,
            fmt,
            linewidth=linewidth,
            alpha=alpha,
            label=label)

    else:
        for row in time_series.data:
            axes.plot(
                time_series.time, row, fmt, linewidth=linewidth, alpha=alpha)
    
    axes.set_ylim([ymin, ymax])


def _add_points_to_axes(axes, time, values):
    """
    Plot circle in the specified ``points``x``values`` for the ``axes``.
    """
    axes.plot(time, values, 'or')


def _add_waveform_to_axes(axes, audio):
    """
    Add the waveform for the time-series `audio` to `axes`.
    """

    # adding the curves
    _add_curve_to_axes(axes, audio, linewidth=None, alpha=0.9)


def _add_waveform_trio_to_axes(axes,
                               audio,
                               rms=None,
                               peak_envelope=None,
                               show_legend=True):
    """
    The waveform trio consists in the curves ``audio``, ``rms`` and
    ``peak_evelope``. This method adds them to ``axes``.
    """
    window_size, hop_size = 2048, 512
    rms = rms or rms_(audio, window_size, hop_size)
    peak_envelope =\
        peak_envelope or peak_envelope_(audio, window_size,
                                                   hop_size)
    # adding the curves
    _add_curve_to_axes(axes, audio, linewidth=None, alpha=0.9)
    _add_curve_to_axes(axes, rms, fmt='r', label=rms.label, set_labels=False)
    _add_curve_to_axes(
        axes,
        peak_envelope,
        fmt='k',
        label=peak_envelope.label,
        set_labels=False)

    if show_legend:
        axes.legend(loc='lower right', ncol=2, fontsize='x-small')


def _add_spectrogram_to_axes(axes,
                             spec,
                             log=False,
                             fmin=0.,
                             fmax=None,
                             normalize=True,
                             cmap='viridis'):
    """
    Add a spectrogram image to the axes
    """

    # handling arguments
    if not fmax:
        fmax = spec.max_frequency
    else:
        if fmax > spec.max_frequency:
            warnings.warn(
                "the fmax specified excceeds the maximum frequency of the spectrum"
            )
            fmax = spec.max_frequency

    freq_indexes = np.logical_and(spec.frequencies >= fmin,
                                  spec.frequencies <= fmax)

    data = spec.data
    if np.any(np.iscomplex(data)):
        data = np.abs(data)
        power = 1.
        db = False
    else:
        power = spec._power
        db = spec._db

    # if the data is not in dB, convert it
    if not db:
        if power == 1.0:
            data = conversion.amplitude_to_db(data)
        elif power == 2.0:
            data = conversion.energy_to_db(data)

    data = data[freq_indexes, :]

    if isinstance(normalize, bool) or normalize is None:
        if normalize:
            vmin, vmax = np.min(data), np.max(data)
        else:
            vmin, vmax = -96, 0
    elif isinstance(normalize, tuple):
        if len(normalize) != 2:
            raise ValueError(
                "If `normalize` is a tuple, it must contain two values (vmin, vmax)."
            )
        vmin, vmax = normalize

    plt.pcolormesh(
        spec.time,
        spec.frequencies[freq_indexes],
        data,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        rasterized=True,
        shading='nearest')

    if log:
        axes.set_yscale("log", basey=2)
        axes.get_yaxis().set_major_formatter(
            matplotlib.ticker.ScalarFormatter())

    axes.set(xlabel='time (s)', ylabel='frequency (Hz)')



"""
iracema is a python package aimed at the extraction of expressive music
information from audio signals
"""
from os.path import dirname as _dirname

import iracema.core.timeseries
import iracema.core.audio
import iracema.core.point
import iracema.core.segment

from iracema.core.timeseries import TimeSeries
from iracema.core.audio import Audio
from iracema.core.point import Point, PointList
from iracema.core.segment import Segment, SegmentList

import iracema.features
import iracema.pitch
import iracema.harmonics
import iracema.segmentation
import iracema.spectral
import iracema.plot
import iracema.util

import iracema.segmentation.onsets

__version__ = u'0.2.1'

root = _dirname(__spec__.origin)



"""
This module contains the implementation of a harmonic extraction model for
monophonic audio.
"""

import numpy as np

from iracema.core.timeseries import TimeSeries
from iracema.util.dsp import n_highest_peaks


def extract(fft, pitch, nharm=16, minf0=24, maxf0=4200, perc_tol=0.04):
    """
    Extract the harmonics of an audio signal, given its `fft` and `pitch`
    curve, using `nharm` harmonics. The pitch detection method will search for
    a a fundamental frequency in the interval between `minf0` and `maxf0`.

    Args
    ----
    fft : FFT
        FFT time series.
    pitch : TimeSeries
        Pitch time series.
    nharm : int
        Number of harmonics to extract.
    minf0 : float
        Minimum fundamental frequency.
    maxf0 : float
        Maximum fundamental frequency.
    min_mag_cand : float
        Minimum amplitude to be candidate, relative to the frame's peak.
    noisiness_thresh : float
        Upper noisiness limit for each frame. If the noisiness exceeds this
        value, the algorithm will consider that the frame has no fundamental
        frequency.
    perc_tol : float
        Frequency tolerance used to search for peaks in the region around the
        expected frequency of each harmonic.

    Returns
    -------
    harmonics : dictionary
        A dictionary containing the results of the harmonics extraction,
        composed of three TimeSeries objects, which can be accessed using the
        following keys: 'frequency', 'magnitude'and 'phase'.
    """
    if minf0 >= maxf0:
        raise ValueError('The parameter maxf0 must be greater than minf0.')
    if nharm < 3:
        raise ValueError('The parameter nharm must be greater than 3.')

    # TODO: handle situation where pitch and fft have different sampling
    # frequencies

    # memory allocation
    harm_mag = np.empty((nharm, fft.nsamples), dtype=np.float_)
    harm_f = np.empty((nharm, fft.nsamples), dtype=np.float_)
    harm_phase = np.empty((nharm, fft.nsamples), dtype=np.float_)

    # run the algorithm for each frame of the fft
    for i, fft_frame in enumerate(np.transpose(fft.data)):
        fft_frame_amp = np.abs(fft_frame)
        fft_frame_phase = np.angle(fft_frame)

        mag, phase, f = frame_harmonics(pitch.data[i],
                                        fft_frame_amp,
                                        fft_frame_phase,
                                        fft.max_frequency,
                                        nharm, perc_tol)

        harm_mag[:, i] = mag
        harm_phase[:, i] = phase
        harm_f[:, i] = f

    harmonics = {}

    harmonics['frequency'] = TimeSeries(
        fft.fs, data=harm_f, start_time=fft.start_time)
    harmonics['magnitude'] = TimeSeries(
        fft.fs, data=harm_mag, start_time=fft.start_time)
    harmonics['phase'] = TimeSeries(
        fft.fs, data=harm_f, start_time=fft.start_time)

    return harmonics


def frame_harmonics(frame_pitch, fft_frame_mag, fft_frame_phase, fft_max_freq,
                    nharm, perc_tol):
    """
    Extract the harmonics for a fft frame.
    """
    N = fft_frame_mag.shape[-1]

    # calculate the index corresponding to the f0
    ix_f0 = int(round(np.nan_to_num(frame_pitch / fft_max_freq * N)))
    delta = ix_f0 * perc_tol

    # memory allocation
    cand_mag = np.zeros((nharm))
    cand_phase = np.zeros((nharm))
    ix_cand_harm = np.zeros((nharm))

    ix_cand_harm[0] = ix_f0
    cand_mag[0] = fft_frame_mag[ix_f0]
    cand_phase[0] = fft_frame_phase[ix_f0]

    # iterate the harmonics
    for j in range(1, nharm):
        ix_f_harm = ix_f0 * (j + 1)

        # check if it is within the boundaries of the FFT
        if ix_f_harm < N:
            # set the tolerance margin to search for the harmonic
            ix_tol_low = int(ix_f_harm - delta)
            ix_tol_hi = int(ix_f_harm + delta)

            # search for the highest peak within the tolerance margin
            val_harm, ix_harm = n_highest_peaks(
                fft_frame_mag[ix_tol_low:ix_tol_hi], 1)
            ix_harm = ix_harm + ix_tol_low

            if ix_harm:
                cand_mag[j] = val_harm
                cand_phase[j] = fft_frame_phase[ix_harm]
                ix_cand_harm[j] = ix_harm
            else:
                cand_mag[j] = fft_frame_mag[(j + 1) * ix_f0]
                cand_phase[j] = fft_frame_phase[(j + 1) * ix_f0]
                ix_cand_harm[j] = (j + 1) * ix_f0
        else:
            # TODO: what if the harmonics are higher than the FFT max
            #       frequency?
            # stop the loop if the frequency index is out of bound
            break

    cand_f = ix_cand_harm * fft_max_freq / N

    return cand_mag, cand_phase, cand_f



import numpy as np
from numpy.lib.stride_tricks import as_strided
from scipy.signal.windows import get_window
from scipy.signal import convolve

import iracema as ir


def convolve_activations(activations, n_frames=5, window_function='hanning'):
    w = get_window(window_function, n_frames, fftbins=False)
    convolved_activations = convolve(activations, w, mode='same')
    convolved_activations = convolved_activations / np.max(convolved_activations)
    return convolved_activations


def three_sliced_mel_spectrograms(audio,
                                  window,
                                  hop,
                                  medium_window_factor=2,
                                  long_window_factor=4,
                                  frames_per_slice=15,
                                  **kwargs):
    """
    Convert audio into a tensor containing three mel spectrograms. 
    
    The tensor contains slices of the spectrograms. Each mel
    spectrogram correspond to one channel.

    Return
    ------
    data : np.array
        Data array of shape (num_slices, frames_per_slice, frequencies,
        channels)
    fs : float
        Sampling frequency of the `data` array.
    """
    medium_window = window * medium_window_factor
    long_window = window * long_window_factor

    mel_spec_long = (ir.spectral.MelSpectrogram(audio, long_window, hop,
                                                **kwargs))
    mel_spec_medium = (ir.spectral.MelSpectrogram(audio, medium_window, hop,
                                                  **kwargs))
    mel_spec_short = (ir.spectral.MelSpectrogram(audio, window, hop,
                                                 **kwargs))

    pad_len = np.floor_divide(frames_per_slice, 2)

    short_len = mel_spec_short.data.shape[1]
    medium_len = mel_spec_medium.data.shape[1]
    long_len = mel_spec_long.data.shape[1]

    dif_long_short = long_len - short_len
    dif_long_medium = long_len - medium_len

    # padding the extremes of the spectrogram
    mel_spec_short = mel_spec_short.pad(pad_len,
                                        pad_len + dif_long_short,
                                        value='repeat')
    mel_spec_medium = mel_spec_medium.pad(pad_len,
                                          pad_len + dif_long_medium,
                                          value='repeat')
    mel_spec_long = mel_spec_long.pad(pad_len, pad_len, value='repeat')

    data = np.stack(
        [mel_spec_short.data, mel_spec_medium.data, mel_spec_long.data])

    # generating a sliced view of the spectrogram
    nchannels, nfrequencies, ntime = data.shape
    nslices = ntime - 2 * pad_len
    new_shape = (nchannels, nfrequencies, frames_per_slice, nslices)
    strides = (nfrequencies * ntime * data.itemsize, ntime * data.itemsize,
               data.itemsize, data.itemsize)
    data = as_strided(data, shape=new_shape, strides=strides)

    data = data.T
    fs = float(mel_spec_short.fs)

    return data, fs


def get_notes_list(onsets, offsets):
    """
    Generate a list of note segments using the specified `onsets` and `offsets`
    arrays.

    Args
    ----
    audio : Audio
        Audio object
    onsets : array
        Indexes of the onset occurrences in `audio`.
    offsets : array
        Indexes of the offset occurrences in `audio`.

    Return
    ------
    notes : list
        List of segments.
    """
    if onsets.shape != offsets.shape:
        raise ValueError("the number of onsets and offsets must the same")

    return [ir.Segment(onsets[i], offsets[i]) for i in range(len(onsets))]




"""
This module contains the implementation of the class TimeSeries.
"""

import copy as cp
from decimal import Decimal

import numpy as np

from iracema.aggregation import sliding_window
from iracema.core.segment import Segment
from iracema.util import conversion
from iracema.util.dsp import but_filter
from iracema.plot import line_plot


class TimeSeries:
    """
    Class for storing and manipulating time series objects, which
    model synchronous discrete time series.

    .. Hint:: This class is also available at the main package level as
        ``iracema.TimeSeries``.

    Attributes
    ----------
    data : numpy array
        Data array sampled at ``fs`` Hz.
    time : numpy array
        Numpy data array containing the time of each sample, relative to the
        original time reference.
    fs : Decimal
        Sampling frequency for the data.
    nyquist : Decimal
        Nyquist frequency for the data.
    ts : Decimal
        Sampling period for the data.
    start_time : Decimal
        The time the time series start (in seconds) relative to the original
        time reference.
    duration : Decimal
        Duration of the time series (in seconds).
    end_time : Decimal
        The time the time series end (in seconds) relative to the original time
        reference.
    unit : str
        String containing the name of the unit for the data (for plotting).
    nsamples : int
        Number of samples per time series in the data array.
    nfeatures
        Number of features for each sample in the data array.
    unit
        Unit for the data (for plotting).
    caption
        Text caption.
    label
        Label for the time series.
    """

    # these properties should be overriden by the subclass if they are supposed
    # to be different
    unit = ''
    caption = ''
    label = ''

    def __init__(self, fs, data=None, start_time=None, unit=None,
                 caption=None):
        """
        Args
        ----
        fs : Decimal
            Sampling frequency for the data.
        data : numpy array, optional
            Data array sampled at ``fs`` Hz. If this argument is not provided,
            the method _write_data() must be called after the initialization to
            set the data array.
        start_time : float, optional
            The time in seconds the time series start, relative to the original
            time reference.
        unit : str, optional
            Unit name for plotting the data of the time series.
        caption : str, optional
            Text caption for the time series.
        """
        if fs <= 0:
            raise ValueError(
                "the sampling frequency (fs) must be greater than zero")

        self.data = None
        self.fs = Decimal(fs)
        self.start_time = Decimal(0) if start_time is None else Decimal(
            start_time)

        if unit:
            self.unit = unit
        if caption:
            self.caption = caption
        if data is not None:
            self._write_data(data)

    @property
    def nsamples(self):  # pylint: disable=missing-docstring
        return self.data.shape[-1]

    @property
    def nfeatures(self):  # pylint: disable=missing-docstring
        if self.data.ndim == 1:
            return 1
        if self.data.ndim == 2:
            return self.data.shape[-2]

    @property
    def duration(self):  # pylint: disable=missing-docstring
        return Decimal(self.nsamples) / Decimal(self.fs)

    @property
    def nyquist(self):  # pylint: disable=missing-docstring
        return Decimal(self.fs) / Decimal(2)

    @property
    def end_time(self):  # pylint: disable=missing-docstring
        return Decimal(self.start_time) + Decimal(self.duration)

    @property
    def ts(self):  # pylint: disable=missing-docstring
        return Decimal(1) / Decimal(self.fs)

    @property
    def time(self):  # pylint: disable=missing-docstring
        start = Decimal(self.start_time)
        step = Decimal(self.duration) / Decimal(self.nsamples)

        return [start + (t * step) for t in range(0, self.nsamples)]

    def copy(self):
        """
        Return a copy of the time series object (deep copy).
        """
        return cp.deepcopy(self)

    def gain(self, db):
        """
        Apply a gain of ``db`` dB to the time series and return the new object.
        """
        scale_factor = conversion.db_to_amplitude(db)
        new = self.copy()
        new.data = new.data * scale_factor

        return new

    def normalize(self, db=0.0):
        """
        Return a copy of the audio time series, normalized to ``db`` dB.
        """
        new = self.copy()
        new.data = self.data / np.max(self.data)
        new = new.gain(db)

        return new

    def diff(self, n=1):
        "Return the n-th discrete difference for the time series"
        nfeatures = self.nfeatures
        dtype = self.data.dtype
        data = np.reshape(self.data, (nfeatures, -1))

        zero_pre_pad = np.zeros((nfeatures, n), dtype)
        padded_data = np.concatenate((zero_pre_pad, data), axis=-1)
        data_diff = np.diff(padded_data, n, axis=-1)

        if nfeatures == 1:
            data_diff.shape = (self.nsamples, )

        ts = self.copy()
        ts.data = data_diff

        return ts
    

    def sliding_window(self,
                       window_size,
                       hop_size,
                       function=None,
                       window_name=None):
        """
        Use a sliding window to aggregate the data from the time series by
        applying ``function`` to each analysis window. The content of each
        window will be passed as the first argument to the function. Return
        the aggregated data in an array.

        Args
        ----
        window_size: int
            Size of the window.
        hop_size : int
            Number of samples to be skipped between two successive windowing
            operations.
        function : function
            Function to be applied to each window. If no function is specified,
            each window will contain an unaltered excerpt of the time series.
        window_name : str
            Name of the window function to be used. Options are: {"boxcar",
            "triang", "blackman", "hamming", "hann", "bartlett", "flattop",
            "parzen", "bohman", "blackmanharris", "nuttall", "barthann",
            "no_window", None}.
        """
        ts = self.copy()
        ts = sliding_window(
            ts, window_size, hop_size, function=function, window_name=window_name)
        return ts

    def zeros_to_nan(self):
        """
        Converts zeros to np.nan in the data array. Returns a new time series.
        """
        ts = self.copy()
        ts.data[self.data == 0] = np.nan
        return ts

    def hwr(self):
        "Return a half-wave rectified copy of the time series."
        rectified_data = np.clip(self.data, 0, None)

        ts = self.copy()
        ts.data = rectified_data

        return ts

   
    def pad(self, pre, post, value=0.):
        """
        Pad the edges of the time series.

        Args
        ----
        pre : int
            Pre-padding length.
        post : int
            Post-padding length.
        value : float or str
            Value for the padding operation. If a float number is provided,
            this value will be used in the padding. If instead the string
            'repeat' is provided, the values at the edges will be repeated
            in the padding operation.
        """
        new = self.copy()
        first_col = np.expand_dims(new.data[...,0], -1)
        last_col = np.expand_dims(new.data[..., -1], -1)
        if isinstance(value, str):
            if value == 'repeat':
                pre_pad_array = np.repeat(first_col, pre, axis=-1)
                post_pad_array = np.repeat(last_col, post, -1)
            else:
                raise ValueError("Invalid value for argument `value`")
        else:
            pre_pad_array = np.full_like(first_col, value)
            post_pad_array = np.full_like(last_col, value)
            pre_pad_array = np.repeat(pre_pad_array, pre, axis=-1)
            post_pad_array = np.repeat(post_pad_array, post, -1)

        new.data = np.concatenate((pre_pad_array, new.data, post_pad_array), axis=-1)
        new.start_time = new.start_time - new.ts*pre

        return new


    def pad_like(self, timeseries, value=0.):
        """
        Pad the end of the current time series to match the length of
        the given time series.
        """
        if self.fs != timeseries.fs:
            ValueError("The sampling rates of both time series must be equal.")
        if self.nsamples > timeseries.nsamples:
            ValueError("The current time series has more samples than the"
                       "given time series.")
        padding_len = timeseries.nsamples - self.nsamples
        new_ts = self.copy()
        if timeseries.data.ndim == 1:
            padding_array = np.ones(padding_len) * value
        elif timeseries.data.ndim == 2:
            padding_array = np.ones((self.nfeatures, padding_len)) * value
        new_ts.data = np.concatenate((new_ts.data, padding_array), axis=-1)
        return new_ts

    def resample_and_pad_like(self, timeseries, value=0.):
        """
        Resample and pad the end of the current time series to match
        the given time series.
        """
        new_ts = self.resample(timeseries.fs)
        new_ts = new_ts.pad_like(timeseries, value=value)
        return new_ts

    def merge(self, timeseries, unit=None, caption=None, start_time=None):
        """
        Merge two time series. Both time series must have the same length and
        sampling frequency. The attributes ``unit``, ``caption`` and
        ``start_time`` of the resulting time series can be optionally set using
        the method's optional arguments. Otherwise these attributes will be
        equal to the values in the instance on which the method was called
        (``self``).
        """
        new_ts = self.copy()
        if self.fs != timeseries.fs:
            raise ValueError(
                'Incompatible sampling frequencies. Both time series must '
                'have the same sampling frequency.')
        if self.nsamples != timeseries.nsamples:
            raise ValueError(
                'Incompatible number of samples. Both time series must have '
                'the same number of samples.')
        new_ts.unit = unit or self.unit
        new_ts.caption = caption or self.caption
        new_ts.start_time = start_time or self.start_time
        new_ts.data = np.vstack((self.data, timeseries.data))

        return new_ts

    def filter(self,
               critical_frequency,
               filter_type='low_pass',
               filter_order=4):
        """
        Filters the time series using a butterworth digital filter. This is
        a wrapper over ``scipy.signal.butter``.

        Arguments
        ---------
        critical_frequency: float
            The critical frequency of frequencies.
        filter_type: [‘lowpass’, ‘highpass’, ‘bandpass’, ‘bandstop’]
            The type of filter.
        filter_order:
            The order of the filter.
        """
        audio_filtered = self.copy()
        audio_filtered.data = but_filter(
            self.data,
            float(self.fs),
            critical_frequency,
            filter_type=filter_type,
            filter_order=filter_order)
        return audio_filtered

    def plot(self, linewidth=1, alpha=0.9, **kwargs):
        "Plot the time series using matplotlib."
        return line_plot(self, linewidth=linewidth, alpha=alpha, **kwargs)

    def time_to_sample_index(self, time):
        """
        Convert time (in seconds) to the correspoding sample index in the time
        series.
        """
        time = time + self.start_time
        index = round(time * self.fs)
        return index

    def __repr__(self):
        """Representation for TimeSeries object."""
        class_name = self.__class__.__name__
        dim = f"({self.nfeatures}, {self.nsamples})"
        other = f"fs={self.fs}, unit={self.unit}, label={self.label}"
        return f"{class_name}: {dim}, {other}"

    def __getitem__(self, sl):
        """
        Get an excerpt from the time series using slices. Return a new
        TimeSeries object.
        """
        if type(sl) == Segment:
            time_offset = sl.start
            sl = sl.generate_slice(self)
        elif (type(sl) == slice):
            index_start = sl.start or sl.stop
            time_offset = conversion.sample_index_to_seconds(
                index_start, self.fs)
        else:
            raise ValueError("invalid value for slicing operation: must be " +
                             "of type `Segment` or a Python slice")

        sliced_data = self.data[sl]
        ts = self.copy()
        ts.data = sliced_data
        ts.start_time += time_offset  # shift start
        return ts

    def get_samples(self, start, stop):
        """
        Get an excerpt from the time series using sample indexes. Return a new
        TimeSeries object.
        """
        return self[..., start:stop]
    
    def log10(self):
        """
        Calculate the base 10 logarithm of the time series.
        """
        new = self.copy()
        new.data = np.log10(self.data)
        return new

    # Arithmetic, relational and boolean operations
    def __add__(self, other):
        """Add two time series."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data + other.data
        return ts

    def __sub__(self, other):
        """Subtract two time series."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data - other.data
        return ts

    def __mul__(self, other):
        """Multiplicate two time series element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data * other.data
        return ts

    def __truediv__(self, other):
        """Divide two time series element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data / other.data
        return ts

    def __mod__(self, other):
        """Division remainder for two time series taken element-wise."""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data % other.data
        return ts

    def __lt__(self, other):
        """Less than"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data < other.data
        return ts

    def __le__(self, other):
        """Less than or equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data <= other.data
        return ts

    def __gt__(self, other):
        """Greater than"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data > other.data
        return ts

    def __ge__(self, other):
        """Greater than or equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data >= other.data
        return ts

    def __eq__(self, other):
        """Equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data == other.data
        return ts

    def __ne__(self, other):
        """Not equal to"""
        if self.data.shape != other.data.shape:
            raise DimensionalityError("The shape of the time series do not "
                                      "match.")
        ts = self.copy()
        ts.data = self.data != other.data
        return ts

    def __len__(self):
        """Length of the time series -- number of samples."""
        return self.data.shape[-1]

    def _write_data(self, data):
        """
        Write the given data to the data field. This method should be
        implemented by subclasses to control how the data field is initialized.

        Be careful using this method; it modifies the object.
        """
        self.data = data

    def _shift_start(self, seconds):
        """
        Shift the start of the time series by the number of seconds given.

        Be careful using this method; it modifies the object.
        """

        self.start_time += seconds


class DimensionalityError(Exception):
    """
    Exception raised for errors in dimensionality of arays
    """



"""
This module contains the implementation of the class ``Audio``.
"""
from librosa.effects import time_stretch, pitch_shift
import resampy
import numpy as np

from iracema.core.timeseries import TimeSeries
from iracema.io.audiofile import read as _read
from iracema.io import player
from iracema.util import conversion


class Audio(TimeSeries):
    """
    Class for storing and manipulating audio data.

    .. Hint:: This class is also available at the main package level as
        ``iracema.Audio``.
    """

    def __init__(self, fs, data, start_time=None, caption=None):
        """
        Example
        -------
        There are two ways to instantiate an ``Audio`` object: from audio files
        or from `NumPy`_ arrays. To initialize it using an audio file, you must
        use the method ``Audio.load()``, specifying the location from which the
        file must be loaded:

        >>> ir.Audio.load('02 - Clarinet - Long Notes.wav')
        Audio: (1, 1785244), fs=44100.0, unit=amplitude, label=waveform

        Alternatively the location can be specified trough an http URL:

        >>> url = (
        ... 'https://raw.githubusercontent.com/cegeme/iracema-audio/master/03 - '
        ... 'Clarinet - Fast Excerpt.wav'
        ... )
        >>> ir.Audio.load(url)
        Audio: (1, 390060), fs=44100.0, unit=amplitude, label=waveform

        To initialize the object using a NumPy array, you can use the
        initializer of the class ``Audio()``. In this case, two arguments
        are necessary: the sampling frequency ``fs`` of the data and the
        ``data`` array.

        >>> ir.Audio(44100, clarinet_data)
        Audio: (1, 390060), fs=44100.0, unit=amplitude, label=waveform

        .. _NumPy:
           https://numpy.org/doc/

        Args
        ----
        fs : float
            Sampling frequency for the data.
        data : numpy array
            Data array containing the audio samples.
        start_time : Decimal, optional
            The time in seconds the time series start, relative to the original
            time reference.
        unit : str, optional
            Unit name for plotting the data of the time series.
        caption : str, optional
            Text caption for the time series.
        data : np.array, optional
            Data vector containing the audio data to be loaded.
        fs : int, optional
            Sampling frequency of the data.
        caption : str, optional
            Textual description used for plotting and displaying reports about
            the audio excerpt.
        """
        unit = 'amplitude'
        self.label = 'waveform'
        self.filename, self.caption = None, caption
        self.filepath = None

        super(Audio, self).__init__(
            fs, data=data, unit=unit, start_time=start_time, caption=caption)

    @classmethod
    def load(cls, file_location, caption=None):
        """
        Load an audio file into an ``Audio`` object.

        Parameters
        ----------
        file_location : str
            Location from which the file will be loaded. The string might
            contain a path pointing to a local file or an http URL referencing
            a remote file
        caption : str, optional
            Caption for the audio file loaded (optional). If this argument is
            not provided, the base name of the loaded file will be used.

        Return
        ------
        audio : Audio
            An object of the class ``Audio``, containing the data loaded from
            the specified location.
        """
        data, fs, base_name = _read(file_location)
        caption = caption or base_name
        audio = cls(fs, data, caption=caption)
        audio.filename = base_name
        audio.filepath = file_location
        return audio

    def resample(self, new_fs):
        """
        Resample time series to ``new_fs``.
        """
        if self.start_time != 0:
            raise (NotImplementedError(
                'The method resample is implemented only for time series '
                'objects with start_time equal to 0.'))
        if self.fs == new_fs:
            return self
        new = self.copy()
        new.data = resampy.resample(self.data, float(self.fs), float(new_fs))
        new.fs = new_fs

        return new
    
    def pitch_shift(self, n_steps, **kwargs):
        """
        Shift the pitch of the audio time series and return the new object.

        This method is a wrapper over librosa_'s ``pitch_shift`` method.

        .. _librosa: https://librosa.org
        """
        new = self.copy()
        new.data = pitch_shift(new.data, new.fs, n_steps, **kwargs)

        return new

    def time_stretch(self, rate, **kwargs):
        """
        Time-stretch the audio time series and return the new object.

        This method is a wrapper over librosa_'s ``time_stretch`` method.
        """
        new = self.copy()
        new.data = time_stretch(new.data, rate, **kwargs)

        return new

    def add_noise(self, db=-50.):
        """
        Add noise to the audio time series and return the new object.

        Arguments
        ---------
        db : float
            Peak amplitude (in dB) of the generated noise, with an amplitude
            of 1.0 correspoding to 0 dB.
        """
        scale_factor = conversion.db_to_amplitude(db)
        new = self.copy()
        rand = np.random.uniform(-scale_factor,scale_factor,self.nsamples)
        new.data = self.data + (rand)

        return new

    def plot(self, linewidth=0.1, alpha=0.9, **kwargs):
        """
        Plot the time series. The ``linewidth`` and ``alpha`` (opacity) values
        can be set using keyword arguments.
        """
        return super(Audio, self).plot(
            linewidth=linewidth, alpha=alpha, **kwargs)

    def play(self):
        """
        Play audio from Audio object.
        """
        return player.play(self)

    def play_with_clicks(self, points):
        """
        Play audio mixed with click sounds in the instants corresponding
        to ``points``.
        """
        return player.play_with_clicks(self, points)

    def mix_clicks(self, points):
        """
        Return an audio time series with clicks mixed in the instants
        corresponding to ``points``.
        """
        return player.play_with_clicks(self, points, return_time_series=True)

    def play_from_time(self, from_time):
        """
        Play audio from Audio object start at time ``from_time``.
        """
        return player.play_interval_seconds(self, from_time, None)

    def play_segment(self, segment):
        """
        Play segment from Audio obejct.
        """
        return player.play_interval_seconds(self, segment.start_time,
                                            segment.end_time)

    @staticmethod
    def stop():
        """
        Stop playing audio.
        """
        player.stop()



"""
Some aggregation methods for time series.
"""
from decimal import Decimal

import numpy as np

from iracema.util.windowing import apply_sliding_window
import iracema.core.timeseries


def sliding_window(time_series, window_size, hop_size, function=None,
                   window_name=None):
    """
    Use a sliding window to aggregate the data from ``time_series`` by applying
    the ``function`` to each analysis window. The content of each window will
    be passed as the first argument to the function. Return the aggregated data
    in an array.

    Args
    ----
    time_series : TimeSeries
        Time series over which the sliding operation must be applied.
    window_size: int
        Size of the window.
    hop_size : int
        Number of samples to be skipped between two successive windowing
        operations.
    function : function
        Function to be applied to each window. If no function is specified,
        each window will contain an unaltered excerpt of the time series.
    window_name : str
        Name of the window function to be used. Options are: {"boxcar",
        "triang", "blackman", "hamming", "hann", "bartlett", "flattop",
        "parzen", "bohman", "blackmanharris", "nuttall", "barthann",
        "no_window", None}.
    """
    def identity(x):
        return x

    if not function:
        function = identity

    new_data = apply_sliding_window(time_series.data, window_size, hop_size,
                                    function, window_name)

    # new sampling frequency for the aggregated time-series
    new_fs = Decimal(time_series.fs) / Decimal(hop_size)

    new_ts = iracema.core.timeseries.TimeSeries(
        new_fs,
        data=new_data,
        start_time=time_series.start_time)

    return new_ts


def aggregate_features(time_series, func):
    """
    Aggregate the features within each sample from ``time_series``.
    """
    new_data = np.apply_along_axis(func, 0, time_series.data)

    new_ts = iracema.core.timeseries.TimeSeries(
        time_series.fs,
        data=new_data,
        start_time=time_series.start_time)

    return new_ts


def aggregate_sucessive_samples(time_series, func, padding='zeros'):
    """
    Aggregate consecutive samples in ``time_series``, and generate a new time
    series object.

    Args
    ----
    time_series : TimeSeries
    padding : {'zeros', 'same', 'ones'}
    """
    # TODO: this kind of aggregation could be simplified with shifting
    # operations in time series. The looping can probably be avoided.
    data = time_series.data
    nfeatures = time_series.nfeatures
    nsamples = time_series.nsamples
    dtype = time_series.data.dtype

    if padding == 'zeros':
        padding_array = np.zeros(nfeatures, dtype)
    elif padding == 'ones':
        padding_array = np.ones(nfeatures, dtype)
    else:
        padding_array = data[..., 0]

    new_data = np.empty(nsamples, dtype)

    # a padded array will be used as 'previous sample' for the aggregation
    # of the first sample
    new_data[0] = func(padding_array, data[..., 0])

    for i in range(1, nsamples):
        new_data[i] = func(data[..., i - 1], data[..., i])

    new_ts = iracema.core.timeseries.TimeSeries(
        time_series.fs,
        data=new_data,
        start_time=time_series.start_time)

    return new_ts



import codecs
import os.path

from setuptools import setup, find_packages

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='iracema',
    url='http://github.com/cegeme/iracema',
    version=get_version('iracema/__init__.py'),
    author='Tairone Magalhães',
    author_email='taironemagalhaes@gmail.com',
    description='Audio Content Analysis for Research on Musical Expressiveness and Individuality',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    install_requires=[
        'numpy>=1.18.5',
        'scipy>=1.6.1',
        'sounddevice>=0.3.12',
        'audioread>=2.1.8',
        'matplotlib==3.3.4',
        'resampy==0.2.2',
        'Deprecated==1.2.10',
        'librosa==0.8.0',
        'w3lib==1.22.0',
        'tensorflow==2.3.2'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(),
    include_package_data = True,
    package_data = {
        'iracema' : [
            'assets/audio/Click.wav',
            'segmentation/clari-onsets.h5'
        ],
    },
)



"""
Useful methods for synthesizing basic sounds.
"""

import numpy as np

from ..timeseries import Audio


def sinusoid(amplitude, frequency, phase, t, fs, soft_start=False):
    """Generate sine wave according to the arguments provided.

    .. math:: y = a sin(2 \pi f t + \phi)

    Args
    ----
    amplitude : float
        Amplitude of the sinusoid.
    frequency : float
        Frequency of the sinusoid.
    phase : float
        The initial phase shift.
    t : float
        Duration in seconds.
    fs : float
        Sampling frequency.
    soft_start : bool
        If True, a linear smoothing amplitude envelope will be applied to
        the first 0.005 seconds of the generated sinusoid to avoid
        click sounds.
    """

    t = np.linspace(0, t, num=t*fs, dtype=np.float_)
    sine_wave = amplitude * np.sin(2*np.pi*frequency*t + phase)

    if soft_start:
        size_soft_start = int(fs * 0.005)
        soft_start_ar = np.concatenate(
            (np.linspace(0, 1, size_soft_start, dtype=np.float_),
             np.ones(sine_wave.shape[-1] - size_soft_start)))
        sine_wave = soft_start_ar * sine_wave

    ts = Audio(sine_wave, fs)

    return ts

